"""Implementation of the various scripts available from bin.

This is structured this way because we want all the significant codes under a
single directory, for analysis, grepping and unit testing.
"""
__author__ = "Martin Blais <blais@furius.ca>"

