"""Libraries and scripts used to create and maintain Beancount documentation in Google Docs.

The code in here is not part of Beancount's usage per-se, but rather used to
automate some of its documentation tasks.
"""
__author__ = 'Martin Blais <blais@furius.ca>'
